from pypdl import Downloader

link = "https://gamedownloads.rockstargames.com/public/installer/Rockstar-Games-Launcher.exe"
dl = Downloader()
# dl.start(link, "test\\", display=True, multithread=True, segments=2, block=False)
# time.sleep(2)
# dl.stop()
dl.start(link, "test")
# dl.start(link, "test\\", display=True, multithread=False)



import copy
import json
import threading
import time
from pathlib import Path
from typing import Dict, Union
from urllib.parse import unquote, urlparse

import requests

MEGABYTE = 1048576
BLOCKSIZE = 4096
BLOCKS = 1024
CHUNKSIZE = BLOCKSIZE * BLOCKS


def get_filename(url: str, headers: Dict) -> str:
    content_disposition = headers.get("Content-Disposition", None)

    if content_disposition and "filename=" in content_disposition:
        filename_start = content_disposition.index("filename=") + len("filename=")
        filename = content_disposition[filename_start:]  # Get name from headers
        filename = filename.strip('"')
    else:
        filename = urlparse(url).path.split("/")[-1]  # Generate name from url

    return unquote(filename)  # Decode URL encodings


def timestring(sec: int) -> str:
    """
    Converts seconds to a string formatted as HH:MM:SS.
    """
    sec = int(sec)
    m, s = divmod(sec, 60)
    h, m = divmod(m, 60)
    return f"{h:02d}:{m:02d}:{s:02d}"


def to_mb(size_in_bytes: int) -> float:
    return size_in_bytes / MEGABYTE


def create_segment_table(
    url: str, file_path: str, segments: str, size: int, etag: Union[str, bool]
) -> Dict:
    """
    Create a segment table for multi-threaded download.
    """
    segments = 5 if (segments > 5) and (to_mb(size) < 50) else segments
    progress_file = Path(file_path + ".json")

    try:
        progress = json.loads(progress_file.read_text())
        if etag and progress["url"] == url and progress["etag"] == etag:
            segments = progress["segments"]
    except Exception:
        pass

    progress_file.write_text(
        json.dumps(
            {"url": url, "etag": etag, "segments": segments},
            indent=4,
        )
    )

    dic = {"url": url, "segments": segments}
    partition_size = size / segments
    for segment in range(segments):
        start = int(partition_size * segment)
        end = int(partition_size * (segment + 1))
        segment_size = end - start
        if segment != (segments - 1):
            end -= 1  # [0-100, 100-200] -> [0-99, 100-200]
        # No segment_size+=1 for last setgment since final byte is end byte

        dic[segment] = {
            "start": start,
            "end": end,
            "segment_size": segment_size,
            "segment_path": f"{file_path }.{segment}.bin",
        }

    return dic


def combine_files(file_path: str, segments: int) -> None:
    """
    Combine the downloaded file segments into a single file.
    """
    with open(file_path, "wb") as dest:
        for i in range(segments):
            segment_file = f"{file_path }.{i}.bin"
            with open(segment_file, "rb") as src:
                while True:
                    chunk = src.read(CHUNKSIZE)
                    if chunk:
                        dest.write(chunk)
                    else:
                        break
            Path(segment_file).unlink()

    progress_file = Path(file_path + ".json")
    progress_file.unlink()


class Basicdown:
    """
    Base downloader class.
    """

    def __init__(self, stop: threading.Event, error: threading.Event):
        self.curr = 0  # Downloaded size in bytes (current size)
        self.completed = 0
        self.id = 0
        self.stop = stop
        self.error = error

    def download(self, url: str, path: str, mode: str, **kwargs) -> None:
        """
        Download data in chunks.
        """
        try:
            with open(path, mode) as f:
                with requests.get(url, stream=True, **kwargs) as r:
                    for chunk in r.iter_content(MEGABYTE):
                        if chunk:
                            f.write(chunk)
                            self.curr += len(chunk)
                        if not chunk or self.stop.is_set() or self.error.is_set():
                            break
        except Exception as e:
            self.error.set()
            time.sleep(1)
            print(f"Error in thread {self.id}: ({e.__class__.__name__}: {e})")


class Simpledown(Basicdown):
    """
    Class for downloading the whole file in a single segment.
    """

    def __init__(
        self,
        url: str,
        file_path: str,
        stop: threading.Event,
        error: threading.Event,
        **kwargs,
    ):
        super().__init__(stop, error)
        self.url = url
        self.file_path = file_path
        self.kwargs = kwargs

    def worker(self) -> None:
        self.download(self.url, self.file_path, mode="wb", **self.kwargs)
        self.completed = 1


class Multidown(Basicdown):
    """
    Class for downloading a specific segment of the file.
    """

    def __init__(
        self,
        segement_table: Dict,
        segment_id: int,
        stop: threading.Event,
        error: threading.Event,
        **kwargs,
    ):
        super().__init__(stop, error)
        self.segment_id = segment_id
        self.segement_table = segement_table
        self.kwargs = kwargs

    def worker(self) -> None:
        url = self.segement_table["url"]
        segment_path = Path(self.segement_table[self.segment_id]["segment_path"])
        start = self.segement_table[self.segment_id]["start"]
        end = self.segement_table[self.segment_id]["end"]
        size = self.segement_table[self.segment_id]["segment_size"]

        if segment_path.exists():
            downloaded_size = segment_path.stat().st_size
            if downloaded_size > size:
                segment_path.unlink()
            else:
                self.curr = downloaded_size

        if self.curr < size:
            start = start + self.curr
            kwargs = copy.deepcopy(self.kwargs)  # since used by others
            kwargs.setdefault("headers", {}).update({"range": f"bytes={start}-{end}"})
            self.download(url, segment_path, "ab", **kwargs)

        if self.curr == size:
            self.completed = 1



# def func(**kwargs):
#     kwargs = kwargs or {x: x for x in range(10)}
#     print(kwargs)


# func()

x = {"x": {1: 2}, "y": {2: 3}}
y = x.copy()
print("x: ", x)
print("y: ", y)
del y["x"]
print("del in y")
print("x: ", x)
print("y: ", y)
del x["y"]
print("del in x")
print("x: ", x)
print("y: ", y)



import os
import threading
import time
from collections import deque
from math import inf
from typing import Callable, Optional

import requests
from reprint import output

from .utls import (
    Multidown,
    Simpledown,
    combine_files,
    create_segment_table,
    get_filename,
    timestring,
    to_mb,
)


class Downloader:
    """
    A multi-threaded file downloader that supports progress tracking, retries, pause/resume functionality etc.

    Keyword Arguments:
        params (dict, optional): A dictionary, list of tuples or bytes to send as a query string. Default is None.
        allow_redirects (bool, optional): A Boolean to enable/disable redirection. Default is True.
        auth (tuple, optional): A tuple to enable a certain HTTP authentication. Default is None.
        cert (str or tuple, optional): A String or Tuple specifying a cert file or key. Default is None.
        cookies (dict, optional): A dictionary of cookies to send to the specified url. Default is None.
        headers (dict, optional): A dictionary of HTTP headers to send to the specified url. Default is None.
        proxies (dict, optional): A dictionary of the protocol to the proxy url. Default is None.
        timeout (number or tuple, optional): A number, or a tuple, indicating how many seconds to wait for the client to make a connection and/or send a response. Default is 20 seconds.
        verify (bool or str, optional): A Boolean or a String indication to verify the servers TLS certificate or not. Default is True.
    """

    def __init__(self, **kwargs):
        self._segement_table = {}
        self._workers = []
        self._threads = []
        self._error = threading.Event()
        self._stop = threading.Event()
        self._kwargs = kwargs or {
            "timeout": 20,
            "allow_redirects": True,
        }  # kwargs that are passed to request module

        self.size = inf
        self.progress = 0
        self.speed = 0
        self.time_spent = 0
        self.downloaded = 0
        self.eta = "99:59:59"
        self.remaining = 0
        self.failed = False
        self.completed = False

    def _display(self, multithread, interval):
        download_mode = "Multi-Threaded" if multithread else "Single-Threaded"
        with output(initial_len=2, interval=interval) as dynamic_print:
            while True:
                if self.size != inf:
                    progress_bar = f"[{'█' * self.progress}{'·' * (100 - self.progress)}] {self.progress}%"
                    progress_stats = f"Total: {to_mb(self.size):.2f} MB, Download Mode: {download_mode}, Speed: {self.speed:.2f} MB/s, ETA: {self.eta}"
                    dynamic_print[0] = progress_bar
                    dynamic_print[1] = progress_stats
                else:
                    download_stats = f"Downloaded: {to_mb(self.downloaded):.2f} MB, Download Mode: {download_mode}, Speed: {self.speed:.2f} MB/s"
                    dynamic_print[0] = "Downloading..."
                    dynamic_print[1] = download_stats

                if self._stop.is_set() or self._error.is_set() or self.completed:
                    break

                time.sleep(interval)

        print(f"Time elapsed: {timestring(self.time_spent)}")

    def _calc_values(self, recent_queue, interval):
        self.downloaded = sum(worker.curr for worker in self._workers)
        self.progress = (
            int(100 * self.downloaded / self.size) if self.size != inf else 0
        )

        # Speed calculation
        recent_queue.appendleft(self.downloaded)
        non_zero_list = [value for value in recent_queue if value]
        if len(non_zero_list) < 2:
            self.speed = 0
        else:
            diff = [a - b for a, b in zip(non_zero_list, non_zero_list[1:])]
            self.speed = to_mb(sum(diff) / len(diff)) / interval

        self.remaining = to_mb(self.size - self.downloaded)
        if self.size != inf and self.speed != 0:
            self.eta = timestring(self.remaining / self.speed)
        else:
            self.eta = "99:59:59"

    def _downloader(self, url, file_path, segments, display, multithread, etag):
        start_time = time.time()

        kwargs = self._kwargs.copy()
        kwargs.pop("params", None)
        head = requests.head(url, **kwargs)

        if head.status_code != 200:
            self._error.set()
            print(f"Server Returned: {head.reason}({head.status_code}), Invalid URL")
            return

        header = head.headers
        filename = get_filename(url, header)
        file_path = file_path or filename

        if os.path.isdir(file_path):
            file_path = os.path.join(file_path, filename)

        if size := int(header.get("content-length", 0)):
            self.size = size

        if not multithread or self.size is inf or header.get("accept-ranges") is None:
            multithread = False
            sd = Simpledown(url, file_path, self._stop, self._error, **self._kwargs)
            th = threading.Thread(target=sd.worker)
            self._workers.append(sd)
            th.start()
        else:
            if etag := header.get("etag", not etag):
                etag = etag.strip('"')

            self._segement_table = create_segment_table(
                url, file_path, segments, self.size, etag
            )
            segments = self._segement_table["segments"]
            for segment in range(segments):
                md = Multidown(
                    self._segement_table,
                    segment,
                    self._stop,
                    self._error,
                    **self._kwargs,
                )
                th = threading.Thread(target=md.worker)
                th.start()
                self._threads.append(th)
                self._workers.append(md)

        interval = 0.15

        if display:
            display_thread = threading.Thread(
                target=self._display, args=[multithread, interval]
            )
            display_thread.start()

        recent_queue = deque([0] * 12, maxlen=12)

        while True:
            status = sum(worker.completed for worker in self._workers)
            self._calc_values(recent_queue, interval)

            if self._stop.is_set() or self._error.is_set():
                break

            if status == len(self._workers):
                if multithread:
                    combine_files(file_path, segments)
                self.completed = True
                break
            time.sleep(interval)

        self.time_spent = time.time() - start_time

    def stop(self):
        """
        Stop the download process.
        """
        time.sleep(3)
        self._stop.set()
        for thread in self._threads:  # waiting for threads to die
            thread.join()

    def start(
        self,
        url: str,
        file_path: Optional[str] = None,
        segments: int = 10,
        display: bool = True,
        multithread: bool = True,
        block: bool = True,
        retries: int = 0,
        mirror_func: Optional[Callable[[], str]] = None,
        etag: bool = True,
    ):
        """
        Start the download process.

        Parameters:
            url (str): The download URL.
            file_path  (str, Optional): The optional file path to save the download. by default it uses the present working directory,
                If file_path  is a directory then the file is downloaded into it else the file is downloaded with the given name.
            segments (int, Optional): The number of segments the file should be divided in multi-threaded download.
            display (bool, Optional): Whether to display download progress and other optional messages.
            multithread (bool, Optional): Whether to use multi-threaded download.
            block (bool, Optional): Whether to block until the download is complete.
            retries (int, Optional): The number of times to retry the download in case of an error.
            mirror_func (function, Optional): A function to get a new download URL in case of an error.
            etag (bool, Optional): Whether to validate etag before resuming downloads.
        """

        def download():
            for i in range(retries + 1):
                try:
                    _url = mirror_func() if i > 0 and callable(mirror_func) else url
                    if i > 0 and display:
                        print(f"Retrying... ({i}/{retries})")

                    self.__init__(**self._kwargs)  # for stop/start func
                    self._downloader(
                        _url, file_path, segments, display, multithread, etag
                    )

                    if not self._error.is_set():
                        break

                    time.sleep(3)

                except Exception as e:
                    print(f"Download Error: ({e.__class__.__name__}, {e})")
                    self._error.set()

            if self._error.is_set():
                self.failed = True

        download_thread = threading.Thread(target=download)
        download_thread.start()

        if block:
            download_thread.join()
